/*
 * select pages from a DVI file.
 *
 * Copyright (c) 1999
 *      WATANABE Takeshi        watanabe@komadori.planet.sci.kobe-u.ac.jp
 * Copyright (c) 2002  the xdvik project team
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL ANY AUTHO OF THIS SOFTWARE BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <string.h>
#include "dvi.h"
#include "pagesel.h"
#include "dvisel.h"
#include "util.h"
#include "dvi-init.h"
#include "events.h"
#include "dvi-draw.h"

#if COLOR || defined(HTEX) /* these require a special stack when saving partial DVI files */

typedef enum { IS_A_HREF, IS_A_COLOR } stack_elem_t;

static char *dvips_papersize_special = NULL;
static struct specials_stack_t color_stack = { 0, NULL};
static struct specials_stack_t href_stack = { 0, NULL};

/*
  stacks saved from the previous page (i.e. active at the
  beginning of the page)
 */
static struct specials_stack_t color_save_stack = { 0, NULL };
static struct specials_stack_t href_save_stack = { 0, NULL };

/*
  A global pointer that points to a stack passed to the select_*
  routines, for passing back information about included files
  (to warn user when a DVI file is not self-contained)
 */
static struct specials_stack_t *global_file_list;

static void
push_stack(struct specials_stack_t *stack, const char *content)
{
    stack->items = xrealloc(stack->items, (stack->stack_len + 1) * sizeof *(stack->items));
    stack->items[stack->stack_len].content = xstrdup(content);
    stack->stack_len++;
}

static Boolean
stack_contains_item(struct specials_stack_t *stack, const char *str)
{
    size_t i;
    for (i = 0; i < stack->stack_len; i++) {
	if (strcmp(str, stack->items[i].content) == 0) {
	    return True;
	}
    }
    return False;
}

static void
pop_stack(struct specials_stack_t *stack)
{
    ASSERT(stack->stack_len >= 1);
    free(stack->items[stack->stack_len - 1].content);
    stack->stack_len--;
}

static void
empty_stack(struct specials_stack_t *stack)
{
    while(stack->stack_len > 0) {
	pop_stack(stack);
    }
}
#endif /* COLOR || defined(HTEX) */

/*
  This is used instead of struct tn, and struct font* from dvi-init.h,
  since it appears that we don't need all the information from there.
*/
static struct fontinfo {
    long TeXnumber;
    unsigned char info[14];
    char *fontname;
    struct fontinfo *next;
    Boolean used;
} *fontinfo_head;

#define PutFour(num, fp) { \
    putc(((num) >> 24) & 0xff, (fp)); putc(((num) >> 16) & 0xff, (fp)); \
    putc(((num) >> 8) & 0xff, (fp)); putc((num) & 0xff, (fp)); }

#define CopyNum(fin, fout, num) { \
    int m = num; \
    while (--m >= 0) putc(getc(fin), fout); \
    }

static FILE *fin, *fout;
long fout_pos;

static void
FontWrite(struct fontinfo *fontp)
{
    if (fontp->TeXnumber > 0xff) {
	if (fontp->TeXnumber > 0xffff) {
	    if (fontp->TeXnumber > 0xffffff) {
		putc(FNTDEF4, fout);
		putc((fontp->TeXnumber >> 24) & 0xff, fout);
		fout_pos++;
	    }
	    else {
		putc(FNTDEF3, fout);
	    }
	    putc((fontp->TeXnumber >> 16) & 0xff, fout);
	    fout_pos++;
	}
	else {
	    putc(FNTDEF2, fout);
	}
	putc((fontp->TeXnumber >> 8) & 0xff, fout);
	fout_pos++;
    }
    else {
	putc(FNTDEF1, fout);
    }
    putc(fontp->TeXnumber & 0xff, fout);
    fwrite(fontp->info, sizeof(char), 14, fout);
    fwrite(fontp->fontname, sizeof(char), fontp->info[12] + fontp->info[13], fout);
    fout_pos += 2 + 14 + fontp->info[12] + fontp->info[13];
}

/*
  Invoked by spcl_scan; saves file names referenced by special commands
  into the global list `included_files_stack'.
*/
static Boolean
scan_for_included_files(char *special, void *data)
{
    UNUSED(data);
    if (global_file_list == NULL) /* no info needed, e.g. when printing */
	return True;
    
    while (*special == ' ' || *special == '\t')
	++special;

    if (memicmp(special, "psfile=", strlen("psfile=")) == 0) {
	char *b_ptr = special + strlen("psfile=");
	char *e_ptr;
	
	while (*b_ptr == '"')
	    b_ptr++;
	
	if ((e_ptr = strchr(b_ptr, '"')) != NULL) {
	    size_t len = e_ptr - b_ptr;
	    char *tmp = xmalloc(len + 1);
	    memcpy(tmp, b_ptr, len);
	    tmp[len] = '\0';
	    if (!stack_contains_item(global_file_list, tmp)) {
		push_stack(global_file_list, tmp);
	    }
	    free(tmp);
	}
    }
    
    return True; /* dummy */
}

#if COLOR || defined(HTEX)
/*
  Invoked by spcl_scan; saves hyperref and color commands
  into the global stacks `color_stack' and `hyperref_stack'.
  FIXME: In reality char *special is treated as const, but the declaration
  of spcl_scan doesn't allow for this.
*/
static Boolean
stack_save(char *special, void *data)
{
    char *ptr;
    size_t len;

    UNUSED(data);
    while (*special == ' ' || *special == '\t')
	++special;
    
    ptr = special;
    
    if (memicmp(special, "color ", len = strlen("color ")) == 0) {
	special += len;
	while (*special == ' ' || *special == '\t')
	    ++special;
	if (memicmp(special, "push ", 5) == 0) {
	    push_stack(&color_stack, ptr);
	}
	else if (memicmp(special, "pop", 3) == 0) {
	    pop_stack(&color_stack);
	}
    }
    else if (memicmp(special, "html:", len = strlen("html:")) == 0) {
	special += len;
	if (memicmp(special, "<a href=", 8) == 0
	    || memicmp(special, "<a name=", 8) == 0) {
	    push_stack(&href_stack, ptr);
	}
	else if (memicmp(special, "</a>", 4) == 0) {
	    pop_stack(&href_stack);
	}
    }
    else if (memicmp(special, "papersize", len = strlen("papersize")) == 0) {
	free(dvips_papersize_special);
	dvips_papersize_special = xstrdup(special);
    }
    return True; /* dummy */
}

/* write a special to the FILE *fp, updating offset as appropriate. */
static void
write_special(const char *str, FILE *fp, long *offset)
{
    unsigned int len = strlen(str);
    if (len < 256) {
	unsigned char c = len;
	putc(XXX1, fp);
	putc(c, fp);
	*offset += 2;
    }
    else { /* how about xxx2, xxx3? It seems that TeX doesn't use them either? */
	putc(XXX4, fp);
	PutFour(len, fp);
	*offset += 5;
    }
    fputs(str, fp);
    *offset += len;
}

/* dump the `open' commands on the current stack to the FILE *fp */
static void
stack_dump_open_commands(FILE *fp, long *pos)
{
    size_t i;
    TRACE_GUI((stderr, "length of stacks: %d, %d", color_save_stack.stack_len, href_save_stack.stack_len));
    for (i = 0; i < color_save_stack.stack_len; i++) {
	TRACE_GUI((stderr, "dumping: |%s|", color_save_stack.items[i].content));
	write_special(color_save_stack.items[i].content, fp, pos);
    }
    for (i = 0; i < href_save_stack.stack_len; i++) {
	TRACE_GUI((stderr, "dumping: |%s|", href_save_stack.items[i].content));
	write_special(href_save_stack.items[i].content, fp, pos);
    }
    if (dvips_papersize_special != NULL)
	write_special(dvips_papersize_special, fp, pos);
    TRACE_GUI((stderr, "end of dumping open\n"));
}

/* dump `close' commands for all the `open' commands on the
   current stack (not the saved one, but the one active at the end of the page)
   to the FILE *fp (closing tags for hyperref anchors,
   `pop' commands for color specials).
*/
static void
stack_dump_close_commands(FILE *fp, long *pos)
{
    size_t i;

    for (i = color_stack.stack_len; i > 0; i--) {
	TRACE_GUI((stderr, "===== color pop"));
	write_special("color pop", fp, pos);
    }
    for (i = href_stack.stack_len; i > 0; i--) {
	TRACE_GUI((stderr, "===== html:</a>"));
	write_special("html:</a>", fp, pos);
    }
}

static void
scan_page_for_specials(int page, Boolean save_stack, Boolean (*special_proc)())
{
    off_t pos_save;
    ubyte my_scan_buffer[DVI_BUFFER_LEN];

    struct drawinf currinf_save;
    ubyte maxchar_save, currinf_pos_save, currinf_end_save;
    size_t i;
    TRACE_GUI((stderr, "parsing page: %d", page));
    if (save_stack) {
	/*
	  first, copy the current stacks (which reflect the state on the
	  preceding page) to color_save_stack and href_save_stack:
	*/
	empty_stack(&color_save_stack);
	empty_stack(&href_save_stack);
	for (i = 0; i < color_stack.stack_len; i++) {
	    TRACE_GUI((stderr, "saving stack: |%s|", color_stack.items[i].content));
	    push_stack(&color_save_stack, color_stack.items[i].content);
	}
	for (i = 0; i < href_stack.stack_len; i++) {
	    TRACE_GUI((stderr, "saving stack: |%s|", href_stack.items[i].content));
	    push_stack(&href_save_stack, href_stack.items[i].content);
	}
    }
    
    (void)fseek(fin, pageinfo_get_offset(page), 0);

    /*
      the datastructures in dvi-draw are rather weird in requiring
      us to provide a buffer, and point the global currinf.pos and end
      pointers to that buffer, for spcl_scan to be able to scan the file.
      First we try to save all the stuff that might be needed by other
      routines:
    */
    pos_save = save_file_status(&currinf_save, &maxchar_save);
    currinf_pos_save = *(currinf.pos);
    currinf_end_save = *(currinf.end);
    lseek(fileno(dvi_file), pageinfo_get_offset(page), SEEK_SET);

    /* then we point currinf to our own buffer: */
    currinf.pos = currinf.end = my_scan_buffer;
    /*
      finally we may invoke spcl_scan(). I hope we can do without
      the
      (!setjmp(some_env))
      black magic ...
    */
    spcl_scan(special_proc, NULL, False);

    /* now we restore those pesky globals, hoping that nobody
       has stamped upon them in the meantime. */
    restore_file_status(currinf_save, maxchar_save, pos_save);
    currinf.pos = &currinf_pos_save;
    currinf.end = &currinf_end_save;

    /* this is an attempt at including setjmp(). */
/* 	for (;;) { */
/* 	    if (read_events(EV_NOWAIT) & EV_GE_NEWPAGE) */
/* 		break; */
/* 	    page_bak = page; */
/* 	    if (!setjmp(canit_env)) { */
/* 		fprintf(stderr, "current position: %lu; seeking to: %lu\n", */
/* 			(unsigned long)pos_save, */
/* 			(unsigned long)pageinfo_get_offset(page)); */
/* 		(void) fprintf(stderr, "seeking to offset %ld, page %d of file %d\n", */
/* 			       pageinfo_get_offset(page), */
/* 			       page, fileno(dvi_file)); */
/* 		lseek(fileno(dvi_file), pageinfo_get_offset(page), SEEK_SET); */
/* 		spcl_scan(stack_save); */
/* 	    } */
/* 	    else { /\* if interrupted *\/ */
/* 		fprintf(stderr, "========= interrupted!\n"); */
/* 		if (page >= page_bak) */
/* 		    page = page_bak; */
/* 		break; */
/* 	    } */
/* 	    if (page >= current_page) */
/* 		break; */
/* 	    page++; */
/* 	} */
/* 	restore_file_status(currinf_save, maxchar_save, pos_save); */
}
#endif /* COLOR || defined(HTEX) */

static void
WriteDVI(int c)
{
    int i, n;
    struct fontinfo *fontp;

    if (c >= FNTNUM0 && c <= FNT4) {
	if (c >= FNT1)
	    n = num(fin, c - FNT1 + 1);
	else
	    n = c - FNTNUM0;
	for (fontp = fontinfo_head; fontp; fontp = fontp->next)
	    if (n == fontp->TeXnumber)
		break;
	if (fontp && fontp->used == False) {
	    fontp->used = True;
	    FontWrite(fontp);
	}
	putc(c, fout);
	fout_pos++;
	switch (c) {
	case FNT4:
	    putc((n >> 24) & 0xff, fout);
	    fout_pos++;
	case FNT3:
	    putc((n >> 16) & 0xff, fout);
	    fout_pos++;
	case FNT2:
	    putc((n >> 8) & 0xff, fout);
	    fout_pos++;
	case FNT1:
	    putc(n & 0xff, fout);
	    fout_pos++;
	default:
	    break;
	}
    }
    else if (c >= FNTDEF1 && c <= FNTDEF4) {
	n = num(fin, c - FNTDEF1 + 1);
	for (fontp = fontinfo_head; fontp; fontp = fontp->next)
	    if (n == fontp->TeXnumber)
		break;
	if (fontp && fontp->used == False) {
	    fontp->used = True;
	    FontWrite(fontp);
	}
	(void) num(fin, 12);
	(void) num(fin, (int)(one(fin) + one(fin)));
    }
    else {
	putc(c, fout);
	fout_pos++;
	n = 0;
	if (c <= XXX4) {
	    for (i = 0; i < c - XXX1 + 1; i++) {
		int x = one(fin);
		putc(x, fout);
		fout_pos++;
		n = (n << 8) | x;
	    }
	}
	switch (c) {
	case SETRULE:
	case PUTRULE:
	    n += 4;
	    /* fall through */
	    
	case RIGHT4:
	case W4:
	case X4:
	case DOWN4:
	case Y4:
	case Z4:
	    n++;
	    /* fall through */

	case RIGHT3:
	case W3:
	case X3:
	case DOWN3:
	case Y3:
	case Z3:
	    n++;
	    /* fall through */

#if defined(Omega)
	case SET2:
	case PUT2: 
#endif /* Omega || KANJI */
	case RIGHT2:
	case W2:
	case X2:
	case DOWN2:
	case Y2:
	case Z2:
	    n++;
	    /* fall through */

	case SET1:
	case PUT1:
	case RIGHT1:
	case W1:
	case X1:
	case DOWN1:
	case Y1:
	case Z1:
#ifdef PTEX
	case TDIR:
#endif
	    n++;
	    /* fall through */

	case XXX1:
	case XXX2:
	case XXX3:
	case XXX4:
	    CopyNum(fin, fout, n);
	    fout_pos += n;
	    /* fall through */

	default:
	    break;
	}
    }
}

/* public function */
Boolean
select_marked_pages(FILE *fp, struct specials_stack_t *stack)
{
    int c, n, page, pagecnt;
    
    unsigned long offset = 0xffffffff;
    struct fontinfo *fontp;

    global_file_list = stack;

    if ((fin = XFOPEN(dvi_name, "r")) == NULL) {
	XDVI_ERROR((stderr, "Cannot open DVI file: %s", strerror(errno)));
	return False;
    }

    /* get font list */
    fseek(fin, g_postamble_offset, 0);
    (void) num(fin, 1+4*6+2+2);
    fontinfo_head = NULL;
    for (;;) {
	if ((c = one(fin)) < FNTDEF1 || c > FNTDEF4) /* maybe POSTPOST */
	    break;
	fontp = (struct fontinfo *)xmalloc(sizeof(struct fontinfo));
	fontp->TeXnumber = num(fin, c - FNTDEF1 + 1);
	fread(fontp->info, sizeof(char), 14, fin);
	n = fontp->info[12] + fontp->info[13];
	fontp->fontname = xmalloc(n);
	fread(fontp->fontname, sizeof(char), n, fin);
	fontp->next = fontinfo_head;
	fontinfo_head = fontp;
	fontp->used = False;
    }

    /* preamble */
    fseek(fin, 0, 0);
    fout = fp;
    fout_pos = pageinfo_get_offset(0);
    CopyNum(fin, fout, fout_pos);

    /* each page */
    pagecnt = 0;
    for (page = 0; page < total_pages; page++) {
	
#if COLOR || defined(HTEX)
	scan_page_for_specials(page, True, stack_save);
#endif
	
	if (!pageinfo_is_marked(page)) {
	    continue;
	}
	else {
	    TRACE_GUI((stderr, "selecting page %d\n", page));
	}
	
	/* Scan the page for included files/graphics, for displaying
	   eventually a warning. Note that this command mustn't
	   change the save stack.
	*/
	scan_page_for_specials(page, False, scan_for_included_files);
	
	CopyNum(fin, fout, 1+4+9*4); /* BOP etc. */
	(void) four(fin);
	PutFour(offset, fout);
	offset = fout_pos;
	fout_pos += 1+4+9*4+4;

#if COLOR || defined(HTEX)
	stack_dump_open_commands(fout, &fout_pos);
#endif
	
	while ((c = getc(fin)) != EOF) {
	    if (! (c & 0x80)) { /* nornal character */
		putc(c, fout);
		fout_pos++;
	    }
	    else if (c == EOP) { /* End Of Page */
#if COLOR || defined(HTEX)
		stack_dump_close_commands(fout, &fout_pos);
#endif
		putc(c, fout);
		fout_pos++;
		break;
	    }
	    else
		WriteDVI(c);
	}
	pagecnt++;
    }

    /* postamble */
    fseek(fin, g_postamble_offset, 0);
    putc(one(fin), fout); /* POST */
    (void) four(fin);
    PutFour(offset, fout);
    offset = fout_pos;
    CopyNum(fin, fout, 4*5+2);
    (void) two(fin);
    putc((pagecnt >> 8) & 0xff, fout);
    putc(pagecnt & 0xff, fout);
    fout_pos += 1+4+4*5+2+2;

    /* ouput font list */
    for (;;) {
	if ((c = one(fin)) < FNTDEF1 || c > FNTDEF4) /* maybe POSTPOST */
	    break;
	n = num(fin, c - FNTDEF1 + 1);
	for (fontp = fontinfo_head; fontp; fontp = fontp->next)
	    if (n == fontp->TeXnumber)
		break;
	if (fontp && fontp->used == True)
	    FontWrite(fontp);
	(void) num(fin, 12);
	(void) num(fin, (int)(one(fin) + one(fin)));
    }

    /* free list */
    for (fontp = fontinfo_head; fontp;) {
	struct fontinfo *nextp;
	free(fontp->fontname);
	nextp = fontp->next;
	free(fontp);
	fontp = nextp;
    }

    /* POSTPOST */
    putc(c, fout);
    (void) four(fin);
    PutFour(offset, fout);
    CopyNum(fin, fout, 1+4);
    for (fout_pos += 1+4+1+4; fout_pos & 3; fout_pos++)
	putc(TRAILER, fout);

    fclose(fin);
    return True;
}

/* public function - FIXME: put common parts of this and previous one into separate function! */
Boolean
select_pagerange(FILE *fp, int from, int to, struct specials_stack_t *stack)
{
    int c, n, page, pagecnt;
    static Boolean written_pre_cmds = False;
    
    unsigned long offset = 0xffffffff;
    struct fontinfo *fontp;

    global_file_list = stack;

/*     fprintf(stderr, "select_pagerange called: %d - %d\n", from, to); */
    if ((fin = XFOPEN(dvi_name, "r")) == NULL) {
	XDVI_ERROR((stderr, "Cannot open DVI file: %s", strerror(errno)));
	return False;
    }

    /* get font list */
    fseek(fin, g_postamble_offset, 0);
    (void) num(fin, 1+4*6+2+2);
    fontinfo_head = NULL;
    for (;;) {
	if ((c = one(fin)) < FNTDEF1 || c > FNTDEF4) /* maybe POSTPOST */
	    break;
	fontp = (struct fontinfo *) xmalloc(sizeof(struct fontinfo));
	fontp->TeXnumber = num(fin, c - FNTDEF1 + 1);
	fread(fontp->info, sizeof(char), 14, fin);
	n = fontp->info[12] + fontp->info[13];
	fontp->fontname = xmalloc(n);
	fread(fontp->fontname, sizeof(char), n, fin);
	fontp->next = fontinfo_head;
	fontinfo_head = fontp;
	fontp->used = False;
    }

    /* preamble */
    fseek(fin, 0, 0);
    fout = fp;
    fout_pos = pageinfo_get_offset(0);
    CopyNum(fin, fout, fout_pos);

    /* each page */
    pagecnt = 0;
    
    for (page = 0; page < total_pages; page++) {
	if (page < from || page > to) {
#if COLOR || defined(HTEX)
	    scan_page_for_specials(page, True, stack_save);
#endif
	}

	if (page >= from && page <= to) { /* it's a selected page */
	    scan_page_for_specials(page, False, scan_for_included_files); 
	    CopyNum(fin, fout, 1+4+9*4); /* BOP etc. */
	    (void) four(fin);
	    PutFour(offset, fout);
	    offset = fout_pos;
	    fout_pos += 1+4+9*4+4;
	    
#if COLOR || defined(HTEX)
	    if (!written_pre_cmds) {
		stack_dump_open_commands(fout, &fout_pos);
		written_pre_cmds = True;
	    }
#endif
	
	    while ((c = getc(fin)) != EOF) {
		if (! (c & 0x80)) { /* nornal character */
		    putc(c, fout);
		    fout_pos++;
		}
		else if (c == EOP) { /* End Of Page */
#if COLOR || defined(HTEX)
		    if (page == to) { /* print close stack for last page */
			stack_dump_close_commands(fout, &fout_pos);
		    }
#endif
		    putc(c, fout);
		    fout_pos++;
		    break;
		}
		else
		    WriteDVI(c);
	    }
	    pagecnt++;
	}
    }

    /* postamble */
    fseek(fin, g_postamble_offset, 0);
    putc(one(fin), fout); /* POST */
    (void) four(fin);
    PutFour(offset, fout);
    offset = fout_pos;
    CopyNum(fin, fout, 4*5+2);
    (void) two(fin);
    putc((pagecnt >> 8) & 0xff, fout);
    putc(pagecnt & 0xff, fout);
    fout_pos += 1+4+4*5+2+2;

    /* output font list */
    for (;;) {
	if ((c = one(fin)) < FNTDEF1 || c > FNTDEF4) /* maybe POSTPOST */
	    break;
	n = num(fin, c - FNTDEF1 + 1);
	for (fontp = fontinfo_head; fontp; fontp = fontp->next)
	    if (n == fontp->TeXnumber)
		break;
	if (fontp && fontp->used == True)
	    FontWrite(fontp);
	(void) num(fin, 12);
	(void) num(fin, (int)(one(fin) + one(fin)));
    }

    /* free list */
    for (fontp = fontinfo_head; fontp;) {
	struct fontinfo *nextp;
	free(fontp->fontname);
	nextp = fontp->next;
	free(fontp);
	fontp = nextp;
    }

    /* POSTPOST */
    putc(c, fout);
    (void) four(fin);
    PutFour(offset, fout);
    CopyNum(fin, fout, 1+4);
    for (fout_pos += 1+4+1+4; fout_pos & 3; fout_pos++)
	putc(TRAILER, fout);

    fclose(fin);
    return True;
}


/*
  Copy entire dvi file to fp.

  This used to be a simple copying loop, but to save info about
  included files to warn user when a DVI file is not self-contained
  (e.g. has included EPS files), we now invoke select_pagerange() which
  can cope with this info.
 */
Boolean
copy_file(FILE *fp, struct specials_stack_t *stack)
{
    ASSERT(fp != NULL);
    return select_pagerange(fp, 0, total_pages, stack);
}


